"""The websocket feed provides real-time updates on orders and trades. These
updates can be applied on to a level 2 or 3 order book snapshot to maintain an 
accurate and up-to-date copy of the exchange order book.

Find more here: `<https://docs.exchange.coinbase.com/#real-time-order-book>`_

.. module:: orderbook
   :synopsis: Real-Time Order Book

.. moduleauthor:: Alexander Simeonov <agsimeon@buffalo.edu>

"""
from json import dumps
from threading import Thread
from time import sleep

from cbexchange.market import MarketClient
from cbexchange.websock import WSClient


class OrderBook(object):
  """`<https://docs.exchange.coinbase.com/#real-time-order-book>`_

  Includes a real-time thread that makes sure the order book is up to date.
  Supports the 'with' statment.

  The order book format will be:

  >>> {
  >>>    "asks": {
  >>>      "da863862-25f4-4868-ac41-005d11ab0a5f": {
  >>>        "price": "295.97",
  >>>        "size": "5.72036512"
  >>>      },
  >>>      ...
  >>>    },
  >>>    "bids": {
  >>>      "3b0f1225-7f84-490b-a29f-0faef9de823a": {
  >>>        "price": "295.96",
  >>>        "size": "0.05088265"
  >>>      },
  >>>      ...
  >>>    },
  >>>    "sequence": 3
  >>> }

  :param str ws_uri:  WebSocket URI.
  :param str api_url:  Coinbase Exchange REST API URI.
  :param str product_id: the product of interest

  """
  WS_URI = 'wss://ws-feed.exchange.coinbase.com'
  API_URI = 'https://api.exchange.coinbase.com'
  PRODUCT_ID = 'BTC-USD'

  def __init__(self, ws_uri=None, api_uri=None, product_id=None):
    self.API_URI = api_uri or self.API_URI
    self.WS_URI = ws_uri or self.WS_URI
    self.PRODUCT_ID = product_id or self.PRODUCT_ID
    self.ws_client = WSClient(ws_uri=self.WS_URI, ws_product_id=self.PRODUCT_ID)
    self.ws_client.connect()
    client = MarketClient(api_uri=self.API_URI)
    self.book = client.get_product_order_book(3, self.PRODUCT_ID)
    self.sequence = self.book['sequence']
    self.die = False
    self.pause = False

    asks = {}
    for entry in self.book['asks']:
      asks[entry[2]] = {'price':entry[0], 'size':entry[1]}
    self.book['asks'] = asks

    bids = {}
    for entry in self.book['bids']:
      bids[entry[2]] = {'price':entry[0], 'size':entry[1]}
    self.book['bids'] = bids

    Thread(target=self._real_time_thread, args=[]).start()

  def _get_key(self, message):
    if message['side'] == 'buy':
      return 'bids'
    else:
      return 'asks'

  def _get_order_id(self, message):
    return message['order_id']

  def _handle_open(self, message):
    key = self._get_key(message)
    self.book[key][self._get_order_id(message)] = {
      'price':message['price'],
      'size':message['remaining_size']
    }

  def _handle_match(self, message):
    maker = message['maker_order_id']
    key = self._get_key(message)
    if maker in self.book[key]:
      left = float(self.book[key][maker]['size'])
      right = float(message['size'])
      self.book[key][maker]['size'] = str(left - right)

  def _handle_done(self, message):
    if message['order_type'] == 'market':
      return
    key = self._get_key(message)
    self.book[key].pop(self._get_order_id(message), None)

  def _handle_change(self, message):
    if message['price'] is None: # market
      return
    if 'new_funds' in message:   # market
      return
    key = self._get_key(message)
    order_id = self._get_order_id(message)
    if order_id in self.book[key]:
      self.book[key][order_id]['size'] = message['remaining_size']

  def _real_time_thread(self):
    """Handles real-time updates to the order book."""
    while self.ws_client.connected():
      if self.die:
        break
      
      if self.pause:
        sleep(5)
        continue

      message = self.ws_client.receive()

      if message is None:
        break

      message_type = message['type']

      if message_type  == 'error':
        continue
      if message['sequence'] <= self.sequence:
        continue

      if message_type == 'open':
        self._handle_open(message)
      elif message_type == 'match':
        self._handle_match(message)
      elif message_type == 'done':
        self._handle_done(message)
      elif message_type == 'change':
        self._handle_change(message)
      else:
        continue

    self.ws_client.disconnect()

  def __enter__(self):
    return self
  
  def __exit__(self, type, value, traceback):
    self.end()

  def __str__(self):
    if self.book:
      return dumps(self.book, indent=2, sort_keys=True)
    else:
      return str(self.book)

  def end(self):
    """Makes sure the real-time thread dies and the WebSocket disconnects."""
    self.die = True

  def pause(self):
    """Stops real-time updates until resume is called."""
    self.pause = True
  
  def resume(self):
    """Resumes real-time updates."""
    self.pause = False

  def get_order_book(self):
    """Returns the Real-Time Order Book.

    :returns: the Real-Time Order Book
    :rtype: dict

    """
    return self.book



"""The WebSocket Feed provides real-time market data updates for orders and
trades. 
   
Find more here: `<https://docs.exchange.coinbase.com/#websocket-feed>`_

.. module:: websock
   :synopsis: WebSocket Feed

.. moduleauthor:: Alexander Simeonov <agsimeon@buffalo.edu>

"""
from json import dumps, loads
from threading import Lock, Thread
from time import sleep

from websocket import create_connection

from cbexchange.client import APIClient


class WSClient(APIClient):
  """API Client for Coinbase Exchange WebSocket Feed.

  This class starts in a disconnected state so make sure to connect before
  attempting to receive any messages.  When using the 'with' statement the
  client connects and disconnects automatically.

  Once connected the client starts a thread which keeps the WebSocket alive 
  using periodic pings. There will be only one keep alive thread per client
  instance. If the WebSocket connection is somehow lost, the keep alive thread 
  will clean up and exit.

  The client is iterable over the messages in the feed:

  :Example:

  >>> from cbexchange.websock import WSClient
  >>> client = WSClient()
  >>> client.connect()
  >>> for message in client:
  >>>   print(message)

  The client supports the 'with' statment:

  :Example:

  >>> from cbexchange.websock import WSClient
  >>> with WSClient() as client:
  >>>   print(client.receive())

  :param str ws_uri:  WebSocket URI.
  :param str ws_type: `<https://docs.exchange.coinbase.com/#subscribe>`_
  :param str ws_product_id: `<https://docs.exchange.coinbase.com/#subscribe>`_

  """
  WS_URI = 'wss://ws-feed.exchange.coinbase.com'
  WS_TYPE = 'subscribe'
  WS_PRODUCT_ID = 'BTC-USD'

  def __init__(self, ws_uri=None, ws_type=None, ws_product_id=None):
    self.WS_URI = ws_uri or self.WS_URI
    self.WS_TYPE = ws_type or self.WS_TYPE
    self.WS_PRODUCT_ID = ws_product_id or self.WS_PRODUCT_ID
    self._ws = None
    self._thread = None
    self._lock = Lock()

  def __iter__(self):
    return self

  def __enter__(self):
    self.connect()
    return self

  def __exit__(self, type, value, traceback):
    self.disconnect()

  def __next__(self):
    """Iterator function for Python 3.

    :returns: the next message in the sequence
    :rtype: dict
    :raises StopIteration: if the WebSocket is not connected

    """
    next = self.receive()
    if next:
      return next
    raise StopIteration

  # Iterator function for Python 2.
  next = __next__

  def _format_message(self, message):
    """Makes sure messages are Pythonic.

    :param str message: raw message
    :returns: Pythonic message
    :rtype: dict

    """
    return loads(message)

  def _keep_alive_thread(self):
    """Used exclusively as a thread which keeps the WebSocket alive."""
    while True:
      with self._lock:
        if self.connected():
          self._ws.ping()
        else:
          self.disconnect()
          self._thread = None
          return
      sleep(30)

  def connect(self):
    """Connects and subscribes to the WebSocket Feed."""
    if not self.connected():
      self._ws = create_connection(self.WS_URI)
      message = {
        'type':self.WS_TYPE,
        'product_id':self.WS_PRODUCT_ID
      }
      self._ws.send(dumps(message))

      # There will be only one keep alive thread per client instance
      with self._lock:
        if not self._thread:
          thread = Thread(target=self._keep_alive_thread, args=[])
          thread.start()

  def disconnect(self):
    """Disconnects from the WebSocket Feed."""
    if self.connected():
      self._ws.close()
      self._ws = None

  def receive(self):
    """Receive the next message in the sequence.

    :returns: the next message in the sequence, None if not connected
    :rtype: dict

    """
    if self.connected():
      return self._format_message(self._ws.recv())
    return None

  def connected(self):
    """Checks if we are connected to the WebSocket Feed.

    :returns: True if connected, otherwise False
    :rtype: bool

    """
    if self._ws:
      return self._ws.connected
    return False



"""Private endpoints are available for order management, and account management.
Every private request must be signed using the described authentication scheme.

Find more here: `<https://docs.exchange.coinbase.com/#private>`_

.. module:: private
   :synopsis: Private API

.. moduleauthor:: Alexander Simeonov <agsimeon@buffalo.edu>

"""
from base64 import b64decode
from hashlib import sha256
from hmac import new
from time import time

from requests import delete, get, post
from requests.auth import AuthBase

from cbexchange.client import RESTClient, PaginationClient

# List Orders Status
STATUS_ALL     = 'all'
STATUS_OPEN    = 'open'
STATUS_PENDING = 'pending'



class CoinbaseExchangeAuth(AuthBase):
  """`<https://docs.exchange.coinbase.com/#authentication>`_

  :param str api_key: the API Key
  :param str secret_key: the base64-encoded signature
  :param str passphrase: the passphrase you specified when creating the API key

  """
  def __init__(self, api_key, secret_key, passphrase):
    self.api_key = api_key
    self.secret_key = secret_key
    self.passphrase = passphrase

  def __call__(self, request):
    timestamp = str(time())
    message = timestamp        + \
              request.method   + \
              request.path_url + \
              (request.body or '')
    hmac_key = b64decode(self.secret_key)
    signature = new(hmac_key, message, sha256)
    signature_b64 = signature.digest().encode('base64').rstrip('\n')

    request.headers.update({
      'CB-ACCESS-SIGN':signature_b64,
      'CB-ACCESS-TIMESTAMP':timestamp,
      'CB-ACCESS-KEY':self.api_key,
      'CB-ACCESS-PASSPHRASE':self.passphrase,
      'Content-Type':'application/json'
    })

    return request

class PrivateClient(RESTClient):
  """`<https://docs.exchange.coinbase.com/#authentication>`_

  :param CoinbaseExchangeAuth auth: authentication for the Private API
  :param str api_uri: Coinbase Exchange REST API URI

  """
  def __init__(self, auth, api_uri=None):
    super(PrivateClient, self).__init__(api_uri)
    self.auth = auth

  def _request(self, method, *relative_path_parts, **kwargs):
    """Sends an HTTP request to the REST API and receives the requested data.

    :param str method: HTTP method name
    :param relative_path_parts: the relative paths for the request URI
    :param kwargs: argument keywords
    :returns: requested data
    :raises APIError: for non-2xx responses

    """
    uri = self._create_api_uri(*relative_path_parts)
    if method == 'get':
      response = get(uri, auth=self.auth, params=kwargs.get('params', None))
    elif method == 'post':
      response = post(uri, auth=self.auth, json=kwargs.get('data', None))
    else:
      response = delete(uri, auth=self.auth, json=kwargs.get('data', None))
    return self._handle_response(response).json()

  def list_accounts(self):
    """`<https://docs.exchange.coinbase.com/#selecting-a-timestamp>`_"""
    return self._get('accounts')

  def get_account(self, account_id):
    """`<https://docs.exchange.coinbase.com/#get-an-account>`_"""
    return self._get('accounts', account_id)

  def get_account_history(self, account_id):
    """`<https://docs.exchange.coinbase.com/#get-account-history>`_"""
    return self._get('accounts', account_id, 'ledger')

  def get_holds(self, account_id):
    """`<https://docs.exchange.coinbase.com/#get-holds>`_"""
    return self._get('accounts', account_id, 'holds')

  def _place_order(self,
                   side,
                   product_id='BTC-USD',
                   client_oid=None,
                   type=None,
                   stp=None,
                   price=None,
                   size=None,
                   funds=None,
                   time_in_force=None,
                   cancel_after=None,
                   post_only=None):
    """`<https://docs.exchange.coinbase.com/#orders>`_"""
    data = {
      'side':side,
      'product_id':product_id,
      'client_oid':client_oid,
      'type':type,
      'stp':stp,
      'price':price,
      'size':size,
      'funds':funds,
      'time_in_force':time_in_force,
      'cancel_after':cancel_after,
      'post_only':post_only
    }
    return self._post('orders', data=data)

  def place_limit_order(self,
                        side,
                        price,
                        size,
                        product_id='BTC-USD',
                        client_oid=None,
                        stp=None,
                        time_in_force=None,
                        cancel_after=None,
                        post_only=None):
    """`<https://docs.exchange.coinbase.com/#orders>`_"""
    return self._place_order(side,
                             product_id=product_id,
                             client_oid=client_oid,
                             type='limit',
                             stp=stp,
                             price=price,
                             size=size,
                             time_in_force=time_in_force,
                             cancel_after=cancel_after,
                             post_only=post_only)

  def place_market_order(self,
                         side,
                         product_id='BTC-USD',
                         size=None,
                         funds=None,
                         client_oid=None,
                         stp=None):
    """`<https://docs.exchange.coinbase.com/#orders>`_"""
    return self._place_order(type='market',
                             side=size,
                             product_id=product_id,
                             size=size,
                             funds=funds,
                             client_oid=client_oid,
                             stp=stp)

  def cancel_order(self, order_id):
    """`<https://docs.exchange.coinbase.com/#cancel-an-order>`_"""
    return self._delete('orders', order_id)

  def cancel_all(self, product_id=None):
    """`<https://docs.exchange.coinbase.com/#cancel-all>`_"""
    return self._delete('orders', data={'product_id':product_id})

  def list_orders(self, status=None):
    """`<https://docs.exchange.coinbase.com/#list-orders>`_"""
    return self._get('orders', params={'status':status})

  def get_order(self, order_id):
    """`<https://docs.exchange.coinbase.com/#get-an-order>`_"""
    return self._get('orders', order_id)

  def list_fills(self):
    """`<https://docs.exchange.coinbase.com/#list-fills>`_"""
    return self._get('fills')

  def _deposit_withdraw(self, type, amount, coinbase_account_id):
    """`<https://docs.exchange.coinbase.com/#depositwithdraw>`_"""
    data = {
      'type':type,
      'amount':amount,
      'coinbase_account_id':coinbase_account_id
    }
    return self._post('transfers', data=data)

  def deposit(self, amount, coinbase_account_id):
    """`<https://docs.exchange.coinbase.com/#depositwithdraw>`_"""
    return self._deposit_withdraw('deposit', amount, coinbase_account_id)

  def withdraw(self, amount, coinbase_account_id):
    """`<https://docs.exchange.coinbase.com/#depositwithdraw>`_"""
    return self._deposit_withdraw('withdraw', amount, coinbase_account_id)

  def _new_report(self,
                  type,
                  start_date,
                  end_date,
                  product_id='BTC-USD',
                  account_id=None,
                  format=None,
                  email=None):
    """`<https://docs.exchange.coinbase.com/#create-a-new-report>`_"""
    data = {
      'type':type,
      'start_date':self._format_iso_time(start_date),
      'end_date':self._format_iso_time(end_date),
      'product_id':product_id,
      'account_id':account_id,
      'format':format,
      'email':email
    }
    return self._post('reports', data=data)

  def new_fills_report(self,
                       start_date,
                       end_date,
                       account_id=None,
                       product_id='BTC-USD',
                       format=None,
                       email=None):
    """`<https://docs.exchange.coinbase.com/#create-a-new-report>`_"""
    return self._new_report(start_date,
                            'fills',
                            end_date,
                            account_id,
                            product_id,
                            format,
                            email)

  def new_accounts_report(self,
                          start_date,
                          end_date,
                          account_id,
                          product_id='BTC-USD',
                          format=None,
                          email=None):
    """`<https://docs.exchange.coinbase.com/#create-a-new-report>`_"""
    return self._new_report(start_date,
                            'account',
                            end_date,
                            product_id,
                            account_id,
                            format,
                            email)

  def get_report_status(self, report_id):
    """`<https://docs.exchange.coinbase.com/#get-report-status>`_"""
    return self._get('reports', report_id)

class PrivatePaginationClient(PaginationClient):
  def __init__(self, auth, api_uri=None, before=True, limit=None, cursor=None):
    super(PrivatePaginationClient, self).__init__(api_uri,
                                                  before,
                                                  limit,
                                                  cursor)
    self.auth = auth

  def _request(self, method, *relative_path_parts, **kwargs):
    """Sends an HTTP request to the REST API and receives the requested data.
    Additionally sets up pagination cursors.

    :param str method: HTTP method name
    :param relative_path_parts: the relative paths for the request URI
    :param kwargs: argument keywords
    :returns: requested data
    :raises APIError: for non-2xx responses

    """
    uri = self._create_api_uri(*relative_path_parts)
    if method == 'get':
      response = get(uri, auth=self.auth, params=kwargs.get('params', None))
    elif method == 'post':
      response = post(uri, auth=self.auth, json=kwargs.get('data', None))
    else:
      response = delete(uri, auth=self.auth, json=kwargs.get('data', None))
    self.is_initial = False
    self.before_cursor = response.headers.get('cb-before', None)
    self.after_cursor = response.headers.get('cb-after', None)
    return self._handle_response(response).json()

class ListAccountsPagination(PrivatePaginationClient):
  """`<https://docs.exchange.coinbase.com/#selecting-a-timestamp>`_"""
  def endpoint(self):
    if self._check_next():
      return self._get('accounts')
    else:
      return None

class AccountPagination(PrivatePaginationClient):
  def __init__(self,
               account_id,
               auth,
               api_uri=None,
               before=True,
               limit=None,
               cursor=None):
    super(AccountPagination, self).__init__(auth,
                                            api_uri,
                                            before,
                                            limit,
                                            cursor)
    self.account_id = account_id

class GetAccountHistoryPagination(AccountPagination):
  """`<https://docs.exchange.coinbase.com/#get-account-history>`_"""
  def endpoint(self):
    if self._check_next():
      return self._get('accounts', self.account_id, 'ledger')
    else:
      return None

class GetHoldsPagination(AccountPagination):
  """`<https://docs.exchange.coinbase.com/#get-holds>`_"""
  def endpoint(self):
    if self._check_next():
      return self._get('accounts', self.account_id, 'holds')
    else:
      return None

class ListFillsPagination(PrivatePaginationClient):
  """`<https://docs.exchange.coinbase.com/#list-fills>`_"""
  def endpoint(self):
    if self._check_next():
      return self._get('fills')
    else:
      return None

class CancelAllPagination(PrivatePaginationClient):
  """`<https://docs.exchange.coinbase.com/#cancel-all>`_"""
  def __init__(self,
               product_id,
               auth,
               api_uri=None,
               before=True,
               limit=None,
               cursor=None):
    super(AccountPagination, self).__init__(auth,
                                            api_uri,
                                            before,
                                            limit,
                                            cursor)
    self.product_id = product_id

  def endpoint(self):
    if self._check_next():
      return self._delete('orders', data={'product_id':self.product_id})
    else:
      return None

class ListOrdersPagination(PrivatePaginationClient):
  """`<https://docs.exchange.coinbase.com/#list-orders>`_"""
  def __init__(self,
               auth,
               status=None,
               api_uri=None,
               before=True,
               limit=None,
               cursor=None):
    super(ListOrdersPagination, self).__init__(auth,
                                               api_uri,
                                               before,
                                               limit,
                                               cursor)
    self.status = status

  def endpoint(self):
    if self._check_next():
      return self._get('orders', params={'status':self.status})
    else:
      return None



"""Handles Errors in the Coinbase Exchange Library.

.. module:: error
  :synopsis: Error Handling

.. moduleauthor:: Alexander Simeonov <agsimeon@buffalo.edu>

"""


class CBExchangeError(Exception):
  """All Coinbase Exchange Library related errors extend this class."""

class APIError(CBExchangeError):
  """Unless otherwise stated, errors to bad requests will respond with HTTP 4xx 
  or status codes. The body will also contain a message parameter indicating the
  cause. Your language's http library should be configured to provide message
  bodies for non-2xx requests so that you can read the message field from the
  body. Find more here: `<https://docs.exchange.coinbase.com/#errors>`_

  :param requests.Response response: HTTP error response

  """

  def __init__(self, response):
    self.status_code = response.status_code
    self.message = response.json().get('message')

  def __str__(self):
    error_string = '%s(%i)' % (self.__class__.__name__, self.status_code)
    if self.message:
      error_string += ': %s' % self.message
    return error_string

class BadRequestError(APIError): pass      # 400
class UnauthorizedError(APIError): pass    # 401
class ForbiddenError(APIError): pass       # 403
class NotFoundError(APIError): pass        # 404
class TooManyRequestsError(APIError): pass # 429
class InternalServerError(APIError): pass  # 500

def get_api_error(response):
  """Acquires the correct error for a given response.

  :param requests.Response response: HTTP error response
  :returns: the appropriate error for a given response
  :rtype: APIError

  """
  error_class = _status_code_to_class.get(response.status_code, APIError)
  return error_class(response)

_status_code_to_class = {
  400: BadRequestError,
  401: UnauthorizedError,
  403: ForbiddenError,
  404: NotFoundError,
  429: TooManyRequestsError,
  500: InternalServerError
}



"""The Market Data API is an unauthenticated set of endpoints for retrieving
market data. These endpoints provide snapshots of market data. 

Find more here: `<https://docs.exchange.coinbase.com/#market-data>`_

.. module:: market
   :synopsis: Market Data API

.. moduleauthor:: Alexander Simeonov <agsimeon@buffalo.edu>

"""
from requests import get

from cbexchange.client import RESTClient, PaginationClient

# Levels for Get Product Order Book
LEVEL_BEST = 1
LEVEL_TOP  = 2
LEVEL_FULL = 3


class MarketClient(RESTClient):
  def _request(self, method, *relative_path_parts, **kwargs):
    """Sends an HTTP request to the REST API and receives the requested data.

    :param str method: HTTP method name
    :param relative_path_parts: the relative paths for the request URI
    :param kwargs: argument keywords
    :returns: requested data
    :raises APIError: for non-2xx responses

    """
    uri = self._create_api_uri(*relative_path_parts)
    response = get(uri, params=kwargs.get('params', None))
    return self._handle_response(response).json()

  def get_products(self):
    """`<https://docs.exchange.coinbase.com/#get-products>`_"""
    return self._get('products')

  def get_product_order_book(self, level=None, product_id='BTC-USD'):
    """`<https://docs.exchange.coinbase.com/#get-product-order-book>`_"""
    return self._get('products', product_id, 'book', params={'level':level})

  def get_product_ticker(self, product_id='BTC-USD'):
    """`<https://docs.exchange.coinbase.com/#get-product-ticker>`_"""
    return self._get('products', product_id, 'ticker')

  def get_trades(self, product_id='BTC-USD'):
    """`<https://docs.exchange.coinbase.com/#get-trades>`_"""
    return self._get('products', product_id, 'trades')

  def get_historic_trades(self, start, end, granularity, product_id='BTC-USD'):
    """`<https://docs.exchange.coinbase.com/#get-historic-rates>`_

    :param start: either datetime.datetime or str in ISO 8601
    :param end: either datetime.datetime or str in ISO 8601
    :pram int granularity: desired timeslice in seconds
    :returns: desired data

    """
    params = {
      'start':self._format_iso_time(start),
      'end':self._format_iso_time(end),
      'granularity':granularity
    }
    return self._get('products', product_id, 'candles', params=params)

  def get_stats(self, product_id='BTC-USD'):
    """`<https://docs.exchange.coinbase.com/#get-24hr-stats>`_"""
    return self._get('products', product_id, 'stats')

  def get_currencies(self):
    """`<https://docs.exchange.coinbase.com/#get-currencies>`_"""
    return self._get('currencies')

  def get_time(self):
    """`<https://docs.exchange.coinbase.com/#time>`_"""
    return self._get('time')

class MarketPaginationClient(PaginationClient):
  def _request(self, method, *relative_path_parts, **kwargs):
    """Sends an HTTP request to the REST API and receives the requested data.
    Additionally sets up pagination cursors.

    :param str method: HTTP method name
    :param relative_path_parts: the relative paths for the request URI
    :param kwargs: argument keywords
    :returns: requested data
    :raises APIError: for non-2xx responses

    """
    uri = self._create_api_uri(*relative_path_parts)
    response = get(uri, params=self._get_params(**kwargs))
    self.is_initial = False
    self.before_cursor = response.headers.get('cb-before', None)
    self.after_cursor = response.headers.get('cb-after', None)
    return self._handle_response(response).json()

class GetTradesPagination(MarketPaginationClient):
  """`<https://docs.exchange.coinbase.com/#get-trades>`_"""
  def __init__(self,
               product_id='BTC-USD',
               api_uri=None,
               before=True,
               limit=None,
               cursor=None):
    super(GetTradesPagination, self).__init__(api_uri, before, limit, cursor)
    self.product_id = product_id

  def endpoint(self):
    if self._check_next():
      return self._get('products', self.product_id, 'trades')
    else:
      return None



"""Handles Clients in the Coinbase Exchange Library.

.. module:: client
   :synopsis: Client Handling

.. moduleauthor:: Alexander Simeonov <agsimeon@buffalo.edu>

"""
from builtins import map
from datetime import datetime

from requests.compat import urljoin, quote

from cbexchange.error import get_api_error


class APIClient(object):
  """Base class of all client in the Coinbase Exchange Library."""

class RESTClient(APIClient):
  """Base class of all clients using the REST API.

  :param str api_uri: Coinbase Exchange REST API URI.

  """
  API_URI = 'https://api.exchange.coinbase.com'

  def __init__(self, api_uri=None):
    self.API_URI = api_uri or self.API_URI

  def _create_api_uri(self, *parts):
    """Creates fully qualified endpoint URIs.

    :param parts: the string parts that form the request URI

    """
    return urljoin(self.API_URI, '/'.join(map(quote, parts)))
  
  def _format_iso_time(self, time):
    """Makes sure we have proper ISO 8601 time.

    :param time: either already ISO 8601 a string or datetime.datetime
    :returns: ISO 8601 time
    :rtype: str

    """
    if isinstance(time, str):
      return time
    elif isinstance(time, datetime):
      return time.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
    else:
      return None

  def _handle_response(self, response):
    """Returns the given response or raises an APIError for non-2xx responses.

    :param requests.Response response: HTTP response
    :returns: requested data
    :rtype: requests.Response
    :raises APIError: for non-2xx responses

    """
    if not str(response.status_code).startswith('2'):
      raise get_api_error(response)
    return response

  def _request(self, method, *relative_path_parts, **kwargs):
    """Abstract method - must be overriden."""
    raise NotImplementedError

  def _get(self, *args, **kwargs):
    """Performs HTTP GET requests.

    :param args: arguments
    :param kwargs: argument keywords
    :returns: requested data
    :raises APIError: for non-2xx responses

    """
    return self._request('get', *args, **kwargs)

  def _post(self, *args, **kwargs):
    """Performs HTTP POST requests.

    :param args: arguments
    :param kwargs: argument keywords
    :returns: requested data
    :raises APIError: for non-2xx responses

    """
    return self._request('post', *args, **kwargs)

  def _delete(self, *args, **kwargs):
    """Performs HTTP POST requests.

    :param args: arguments
    :param kwargs: argument keywords
    :returns: requested data
    :raises APIError: for non-2xx responses

    """
    return self._request('post', *args, **kwargs)

class PaginationClient(RESTClient):
  """Handles Pagination `<https://docs.exchange.coinbase.com/#pagination>`_

  This client is iterable in the given direction (before parameter).

  :param str api_uri: Coinbase Exchange REST API URI.
  :param bool before: if True goes to the page before (older) else after (newer)
  :param int limit: number of results per request. Maximum 100. (default 100)
  :param cursor: the cursor you want to start at (default latest)

  """
  def __init__(self, api_uri=None, before=True, limit=None, cursor=None):
    super(PaginationClient, self).__init__(api_uri)
    self.before = before
    self.limit = limit
    self.before_cursor = None
    self.after_cursor = None
    if cursor:
      self.is_initial = False
      if before:
        self.before_cursor = cursor
      else:
        self.after_cursor = cursor
    else:
      self.is_initial = True

  def __iter__(self):
    return self

  def __next__(self):
    """Iterator function for Python 3.

    :returns: the next message in the sequence
    :raises StopIteration: if there are no more messages

    """
    next = self.endpoint()
    if next is not None:
      return next
    raise StopIteration

  # Iterator function for Python 2.
  next = __next__

  def _get_params(self, **kwargs):
    params = kwargs.get('params', {})
    params['limit'] = self.limit
    if not self.is_initial:
      if self.before:
        params['before'] = self.before_cursor
      else:
        params['after'] = self.after_cursor
    return params

  def _check_next(self):
    """Checks if a next message is possible.

    :returns: True if a next message is possible, otherwise False
    :rtype: bool

    """
    if self.is_initial:
      return True
    if self.before:
      if self.before_cursor:
        return True
      else:
        return False
    else:
      if self.after_cursor:
        return True
      else:
        return False

  def is_before(self):
    """Returns True if the direction is set to before.

    :returns: True if the direction is set to before, otherwise False
    :rtype: bool

    """
    return self.before

  def is_after(self):
    """Returns True if the direction is set to after.

    :returns: True if the direction is set to after, otherwise False
    :rtype: bool

    """
    return not self.before

  def set_before(self):
    """Sets the direction to before."""
    self.before = True

  def set_after(self):
    """Sets the direction to after."""
    self.before = False

  def endpoint(self):
    """Abstract method - must be overriden. Performs the endpoint operation"""
    raise NotImplementedError

  def get_before_cursor(self):
    """Acquires the before cursor.

    :returns: the before cursor

    """
    return self.before_cursor

  def get_after_cursor(self):
    """Acquires the after cursor.

    :returns: the after cursor

    """
    return self.after_cursor



from distutils.core import setup
setup(
  name = 'cbexchange',
  packages = ['cbexchange'],
  version = '1.0.1',
  description = 'CBExchange - Coinbase Exchange Python API',
  author = 'Alexander Simeonov',
  author_email = 'agsimeon@buffalo.edu',
  url = 'https://github.com/agsimeonov/cbexchange',
  download_url = 'https://github.com/agsimeonov/cbexchange/tarball/1.0.0',
  keywords = ['cbexchange',
              'coinbase',
              'api',
              'bitcoin',
              'client',
              'exchange',
              'websocket',
              'orderbook',
              'order',
              'book'],
  install_requires=[
    'requests',
    'websocket-client'
  ],
  classifiers = [],
)